html, body {
  margin: 0;
  padding: 0;
  font-family: "Verdana", "Geneva", "sans-serif";
}

#main h1, h3 {
  margin: 0px;
}

#main h3 {
  margin-bottom: 8px;
}


.hash {
  font-size: 12px;
  color: #666;
}

#main .attr {
  display: block;
  margin: 3px 0;

} 

#main .desc {
  font-weight: bold;
}



.hidden {
  display: none;
}


.fold {
  margin-top: 15px;
  font-size: 2em;
}

.show_all {
  color: rgb(50, 38, 41);
  float: right;
  margin-right: 50px;
}

.show_all:hover {
  color: rgb(150, 188, 221);
}

ul.mod-list, ul.env-list, ul.fac-list {
  padding: 0;
  margin: 0;
  list-style-type: none;
}

ul.mod-list li, ul.fac-list li {
  border-left: 3px solid black;
  margin: 6px 4px;
  padding: 1px 3px;
  max-width: 500px;
}


ul.mod-list li div {

}

ul.mod-list li .version, ul.fac-list li .mode, ul.fac-list li .buffering {
  float: right;
  margin-left: 20px;
}

ul.mod-list li .version {
  text-align: left;
}

.clear {
  clear: both;
}

.fold:hover {
  cursor: pointer;
}

ul.env-list li {
  margin: 3px;
  margin-left: 7px;
  font-size: 0.8em;
}

ul.env-list li .key {
  font-weight: bold;
}

