body {
    overflow: hidden;
}

/* Splitter */

#splitter {
  height: 100vh;
  width: 100%;
}

#splitter > div {
  overflow: auto;
}

.hsplitbar {
  width: 5px;
  background: #aaa;
}

/* Trial info */

#side {
  background: #fff;
  font-size: 14px;
}

#side-internal {
  margin: 5px;
}



/* History filter */

.filter .internal {
  margin: 4px 8px;
}

.filter .internal select {
  width: 100%;
}


.now .toolbar {
  margin: 5px;
}

.exec-selection {
    margin-top: 2px;
}

.filter {
    width: 200px;
    float: left;
    border-right: 1px solid black;
    height: 100%;
}

#top {
    overflow: hidden !important;
}